(* Mathematica Package *)

(* Created by the Wolfram Workbench Jul 30, 2012 *)

(* Copyright 2012 by Johnathan Burchill (j.burchill@ucalgary.ca) *)

BeginPackage["MadrigalWeb`"]
(* Exported symbols added here with SymbolName::usage *) 

getInstruments::usage = "getInstruments[madrigalServer] returns information about instruments in the server's database.";
getExperiments::usage = "getExperiments[madrigalServer] returns information about experiments in the server's database.";
getMetaData::usage = "getMetaData[madrigalServer,\"FileType\"->type] returns the table file of the specifed type.";
getExperimentFiles::usage = "getExperimentFiles[madrigalServer] returns information about experiment files.";
getParameters::usage = "getParameters[madrigalServer, filename] returns information about parameters in the specified file.";
getData::usage = "isPrint[madrigalServer, filename, parameterlist, fullName, email, affiliation] returns values for the specified parameters.";
positionCalculator::usage = "positionCalculator[madrigalServer, date, latSpec, lonSpec, altSpec, parameterList] returns derived parameters that depend on location.";
timeCalculator::usage = "timeCalculator[madrigalServer, startDate, endDate, hourStep, parameterList] returns derived parameters that depend only on time.";
radarToGeodetic::usage = "radarToGeodetic[madrigalServer, radarCoords, beamCoords] returns the radar beam coordinates in geodetic frame.";
geodeticToRadar::usage = "geodeticToRadar[madrigalServer, radarCoords, beamCoords] returns the radar beam coordinates in radar frame (azimuth, elevation, range).";
traceMagneticField::usage = "traceMagneticField[madrigalServer, date, inputType, outputType, startingPoints] returns location of traced magnetic field line.";


Begin["`Private`"]
(* Implementation of the package *)

(* getInstrument *)

getInstruments[madServer_String]:=Module[{results},
	results = Import[constructServiceRequest[madServer,"getInstrumentsService.py"],"CSV"];
	(Thread[{"Name","Code","Mnemonic","Latitude","Longitude","Altitude","Type"}->#[[1;;7]]])&/@results
]

(* getExperiments *)
Options[getExperiments] = {"Local"->False,"InstrumentCode"->All};
getExperiments[madrigalServer_String, startDate_,endDate_,opts:OptionsPattern[]]:= Module[{sites,ic,codeq,startdateq,enddateq,operation,results,localq},
	sites = getMetaData[madrigalServer, "FileType"->5][[All,{1,3}]];
    ic = OptionValue["InstrumentCode"];
	codeq=Switch[ic,
		All, StringJoin[("code="<>ToString[#])&/@sites[[All,1]]],
		_List, StringJoin[("code="<>ToString[#])&/@ic],
		_, "code="<>ToString[ic]];
    startdateq=dateQuery[startDate,"start"];
    enddateq=dateQuery[endDate,"end"];
    localq = "local="<>(OptionValue["Local"]/.{False->"0",True->"1"});
	operation = constructServiceRequest[madrigalServer,"getExperimentsService.py",{codeq,startdateq,enddateq,localq}];			
	results = Import[operation,"CSV"];
	((Thread[{"Id", "Url", "Name", "SiteId", "SiteName", "InstCode", "InstName", "Local", "PI", "ContactEmail"} -> 
		Delete[#, List /@ Range[8, 19]]]~Join~Thread[{"StartDate","EndDate"}->Partition[#[[8;;19]],6]] )&/@results)/.Rule["Local",v_]:>Rule["Local",v/.{1->True,0->False}]
]

(* getMetaData *)
Options[getMetaData] = {"FileType"->0};
getMetaData[madrigalServer_String, opts:OptionsPattern[]]:=Module[{type},
	type = OptionValue["FileType"];
	If[!MemberQ[Range[0,9],type],Return[]];
	Import[constructServiceRequest[madrigalServer,"getMetadata", "fileType="<>ToString@type]]
]


(* getExperimentFiles *)
Options[getExperimentFiles] = {"GetNonDefault"->False};
getExperimentFiles[madrigalServer_String, experimentId_, opts:OptionsPattern[]]:= Module[{getnondefault,operation,results},
    getnondefault = OptionValue["GetNonDefault"];
	operation = constructServiceRequest[madrigalServer,"getExperimentFilesService.py","id="<>ToString[experimentId]];			
	results = Import[operation,"CSV"];
	Thread[	{"Name","Kindat","KindatDescription","Category",
		"Status","Permission","ExperimentId"}->Append[#,experimentId]]&/@(If[getnondefault === False, results /. {_, _, _, 0, _, _} :> Sequence[]])
]

(* getParameters *)
getParameters[madrigalServer_String, filename_String]:=Module[{bool,operation,results},
	bool:={"0"->False,"1"->True};
	operation = constructServiceRequest[madrigalServer,"getParametersService.py","filename="<>filename];			
	results = Import[operation,"Lines"];
	StringSplit[#, "\\"] & /@ results /. {mnemonic_, desc_, error_, unit_, meas_, cat_, sure_, add : Repeated[_, {0, 1}]} :> 
		{"Mnemonic" -> mnemonic, "Description" -> desc, "IsError" -> (error /. bool), "Unit" -> unit, 
	     "IsMeasured" -> (meas /. bool), "Category" -> cat, "IsSure" -> (sure /. bool), 
         "IsAddIncrement" -> If[Length@{add} === 0, False, (add /. bool)]}
]

(* getData *)
Options[getData] = {"Filters"->None, "IgnoreMissing"->False};
getData[madrigalServer_String, filename_String, parameters_List, 
	fullName_String, affiliation_String, email_String, opts:OptionsPattern[]]:= 
	Module[{filtersq,ignoremissing,hasDate,operation,paramq,userq,results,datePos,date},
		filtersq = OptionValue["Filters"]/.{None->"",f_:>"filters="<>StringJoin[Riffle[f," "]]};
		ignoremissing = OptionValue["IgnoreMissing"];
		hasDate = MemberQ[parameters, "Date"];
		paramq = constructParameterQuery[parameters];
		userq = constructQueryList[{"user_fullname=","user_affiliation","user_email"}, {fullName, affiliation, email}];
		operation = constructServiceRequest[madrigalServer,"isprintService.py",{"file="<>filename,paramq,filtersq,userq}];
		results = ToExpression[StringSplit/@Import[operation,"Lines"]/.{{""}->Sequence[],"missing"->"Missing[]"}];
		If[hasDate, datePos=Flatten@Position[parameters,"Date"]//First;
			results=(date=DateList[#[[datePos;;datePos+5]]];Insert[Delete[#, List /@ Range[datePos,datePos+5]], date, datePos])&/@results;
		];
		results
	]

(* positionCalculator *)
(* specs: {start, end, step} *)
positionCalculator[madrigalServer_String, date_, latSpec_, lonSpec_, altSpec_, parameters_List]:=
Module[{dateq,latq,lonq,altq,paramq,operation,results},
		dateq=dateQuery[date];
		latq=constructQueryList[{"startLat","endLat","stepLat"},latSpec];
		lonq=constructQueryList[{"startLong","endLong","stepLong"},lonSpec];
		altq=constructQueryList[{"startAlt","endAlt","stepAlt"},altSpec];
		paramq=constructParameterQuery[parameters,","];
		operation=constructServiceRequest[madrigalServer,"madCalculatorService.py",{dateq,latq,lonq,altq,paramq}];
		results = Import[operation, "Table"]
	]

(* timeCalculator *)
timeCalculator[madrigalServer_String, startDate_, endDate_, hourStep_, parameters_List]:=
Module[{startdateq,enddateq,hourstepq,paramq,operation,results},
		startdateq=dateQuery[startDate,"start"];
		enddateq=dateQuery[endDate,"end"];
		hourstepq="stephours="<>ToString[hourStep];
		paramq=constructParameterQuery[parameters,","];
		operation=constructServiceRequest[madrigalServer,"madTimeCalculatorService.py",{startdateq,enddateq,hourstepq,paramq}];
		results = {DateList[#[[1;;6]]],Sequence@@(#[[7;;]])}&/@Import[operation, "Table"]
	]

(* radarToGeodetic *)
(* radarCoords: geodetic {lat, long, alt} *)
(* beamCoords: list of triples of {azimuth, elevation, and range} *)
radarToGeodetic[madrigalServer_String, radarCoords_, beamCoords_]:=
Module[{azimuths,elevations,ranges,azq,elq,rangeq,radarq,operation},
		{azimuths,elevations,ranges}=Transpose[beamCoords];
		azq="az="<>StringJoin@Riffle[ToString/@azimuths,","];
		elq="el="<>StringJoin@Riffle[ToString/@elevations,","];
		rangeq="range="<>StringJoin@Riffle[ToString/@ranges,","];
		radarq=constructQueryList[{"slatgd","slon","saltgd"},radarCoords];
		operation=constructServiceRequest[madrigalServer,"radarToGeodeticService.py",{radarq,azq,elq,rangeq}];
		Import[operation, "CSV"]
	]


(* geodeticToRadar *)
(* radarCoords: geodetic {lat, long, alt} *)
(* beamCoords: list of triples of geodetic {latitude, longitude, altitude} *)
geodeticToRadar[madrigalServer_String, radarCoords_, beamCoords_]:=
Module[{lat,lon,alt,gdlatq,glonq,gdaltq,radarq,operation},
		{lat,lon,alt}=Transpose[beamCoords];
		gdlatq="gdlat="<>StringJoin@Riffle[ToString/@lat,","];
		glonq="glon="<>StringJoin@Riffle[ToString/@lon,","];
		gdaltq="gdalt="<>StringJoin@Riffle[ToString/@alt,","];
		radarq=constructQueryList[{"slatgd","slon","saltgd"},radarCoords];
		operation=constructServiceRequest[madrigalServer,"geodeticToRadarService.py",{radarq,gdlatq,glonq,gdaltq}];
		Import[operation, "CSV"]
	]

(* traceMagneticField *)
(* startPoints: list of triples of geodetic {lat,lon,alt} or GSM {x,y,z} *)
Clear[traceMagneticField];
Options[traceMagneticField] = {"Model"->"Tsyganenko","TargetLocation"->"Conjugate","StopAlt"->1000,"InCoordinates"->"GSM","OutCoordinates"->"GSM"};
traceMagneticField[madrigalServer_String, date_, startPoints_,opts:OptionsPattern[]]:=
Module[{models,targets,coordinates,model,target,stopalt,intype,outtype,dateq,miscq,in1,in2,in3,
	in1q,in2q,in3q,operation,results},
		models={"Tsyganenko"->0,"IGRF"->1};
		targets={"Conjugate"->0,"NorthAlt"->1,"SouthAlt"->2,"Apex"->3,"GSM XY Plane"->4};
		coordinates={"Geodetic"->0,"GSM"->1};
		model=OptionValue["Model"]/.models;
		target=OptionValue["TargetLocation"]/.targets;
		stopalt=OptionValue["StopAlt"];
		intype=OptionValue["InCoordinates"]/.coordinates;
		outtype=OptionValue["OutCoordinates"]/.coordinates;
		If[!MemberQ[Range[0,1],model]||!MemberQ[Range[0,4],target]||
			!MemberQ[Range[0,1],intype]||!MemberQ[Range[0,1],outtype],Return[]];
		dateq=dateQuery[date];
		miscq=constructQueryList[{"model","inputType","outputType","qualifier","stopAlt"},{model,intype,outtype,target,stopalt}];
		{in2,in3,in1}=Transpose[startPoints];
		in1q="in1="<>StringJoin@Riffle[ToString/@in1,","];
		in2q="in2="<>StringJoin@Riffle[ToString/@in2,","];
		in3q="in3="<>StringJoin@Riffle[ToString/@in3,","];
		operation=constructServiceRequest[madrigalServer,"traceMagneticFieldService.py",{dateq,miscq,in1q,in2q,in3q}];
		results = Import[operation, "Data"]/.{"missing"->Missing[]}
	]



(* helper expressions *)
dateQuery[date_,prefix_:""]:= constructQueryList[prefix<>#&/@{"year","month","day","hour","min","sec"},Round/@DateList[date]];
constructServiceRequest[server_String,service_String]:=constructServiceRequest[server,service,{}];
constructServiceRequest[server_String,service_String,queries_]:=server<>"/"<>service<>If[Length@Flatten@{queries}===0,"","?"<>StringJoin[Riffle[Flatten[{queries}],"&"]]]
constructQueryList[names_List, values_List]:=MapThread[StringJoin[#1, "=", #2]&,{names,ToString/@values}];
constructParameterQuery[parameters_,separator_:"+"]:= "parms="<>StringJoin[Riffle[Flatten[parameters/."Date"->{"year","month","day","hour","min","sec"}],separator]];
toBool:={"0"->False,"1"->True};

End[]

EndPackage[]

